use tokio::sync::{mpsc, oneshot};

use super::{
    registry::{BasicMessageType, Registry},
    BrokerResult,
};
use crate::{Message, Publication, Subscription};

pub(crate) trait BrokerRequestInternal {
    fn handle_request(self: Box<Self>, registry: &mut Registry);
}

pub(crate) trait BrokerRequest: BrokerRequestInternal {}
impl<T: BrokerRequestInternal> BrokerRequest for T {}

pub(crate) type BrokerRequestBox = Box<dyn BrokerRequest + Send + Sync>;
pub(crate) type BrokerRequestSender = mpsc::Sender<BrokerRequestBox>;

pub(crate) struct SubscribeRequest<T> {
    msg_type: BasicMessageType<T>,
    response_tx: oneshot::Sender<BrokerResult<Subscription<T>>>,
}

impl<T: Message> BrokerRequestInternal for SubscribeRequest<T> {
    fn handle_request(self: Box<Self>, registry: &mut Registry) {
        let sender = registry.get_sender_for_type::<T>(&self.msg_type);
        let subscription = sender.map(|sender| {
            let receiver = sender.subscribe();
            Subscription::new(receiver)
        });
        let _ = self.response_tx.send(subscription);
    }
}

impl<T: Message> SubscribeRequest<T> {
    pub(crate) fn new() -> (Box<Self>, oneshot::Receiver<BrokerResult<Subscription<T>>>) {
        let (response_tx, response_rx) = oneshot::channel();
        (
            Box::new(Self {
                msg_type: Default::default(),
                response_tx,
            }),
            response_rx,
        )
    }
}

pub(crate) struct AdvertiseRequest<T> {
    msg_type: BasicMessageType<T>,
    response_tx: oneshot::Sender<BrokerResult<Publication<T>>>,
}

impl<T: Message> BrokerRequestInternal for AdvertiseRequest<T> {
    fn handle_request(self: Box<Self>, registry: &mut Registry) {
        let sender = registry.get_sender_for_type::<T>(&self.msg_type);
        let publication = sender.map(|sender| Publication::new(sender.clone()));
        let _ = self.response_tx.send(publication);
    }
}

impl<T: Message> AdvertiseRequest<T> {
    pub(crate) fn new() -> (Box<Self>, oneshot::Receiver<BrokerResult<Publication<T>>>) {
        let (response_tx, response_rx) = oneshot::channel();
        (
            Box::new(Self {
                msg_type: Default::default(),
                response_tx,
            }),
            response_rx,
        )
    }
}
